/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                                                                        **
**                           BIALYSTOK 2007                               **
**                                                                        **
**                       MODULE UDP Protocol Service                      **
**                                                                        **
**                     Mikrokontroler ARM: LPC2378                        **
**                                                                        **
**                     Wersja z dnia : 2007.03.22                         **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/
#include "udpservice.h"
#include "types.h"
#include "ipservice.h"
#include "frameutil.h"
#include "trace.h"


#define MaxUDPSegmentSize  EthMaxFrameSzie - IPHeaderSize - UDPHeaderSize
//      1508 =             1536            - 20           - 8

#define SocketArraySize     8

static UDPSocketDescriptionRecT UDPSocketArray [ SocketArraySize ] ;
static UCHAR UDPSocketSwitch [ SocketArraySize ] ;


static USHORT CheckUnique ( PtrUDPLayerInstanceRecT UDPLayerInstance ,
                            USHORT                  PortNo )
{
  USHORT Loop ;
  PtrUDPSocketDescriptionRecT TempSocket ;
  /*-----------------------------------*/
  for ( Loop = 0 ; Loop < SocketArraySize ; Loop ++ )
  {
    if ( UDPSocketSwitch [ Loop ] )
    {
      TempSocket = & UDPSocketArray [ Loop ] ;
      if ( TempSocket -> ListenPortNo == PortNo )
        return ( FALSE ) ;
    } /* if */ ;
  } /* for */ ;
  return ( TRUE ) ;
} /* CheckUnique */


static USHORT AllocUDPPortN ( PtrUDPLayerInstanceRecT UDPLayerInstance )
{
  USHORT PortNo ;
  USHORT Unique ;
  /*-----------------------------------*/
  for ( ; ; )
  {
    PortNo = UDPLayerInstance -> LastLocalPort ++ ;
    if ( UDPLayerInstance -> LastLocalPort > AutoIncrementUDPPortNoUpBound )
      UDPLayerInstance -> LastLocalPort = AutoIncrementUDPPortNoLoBound ;
    Unique = CheckUnique ( UDPLayerInstance , PortNo ) ;
    if ( Unique )
      return ( PortNo ) ;
  } /* for */ ;
} /* AllocUDPPortN */


PtrUDPSocketDescriptionRecT AllocateUDPSocket ( PtrUDPLayerInstanceRecT UDPLayerInstance )
{
  USHORT Loop ;
  PtrUDPSocketDescriptionRecT NewSocket ;
  /*-----------------------------------*/
  for ( Loop = 0 ; Loop < SocketArraySize ; Loop ++ )
  {
    if ( ! UDPSocketSwitch [ Loop ] )
    {
      UDPSocketSwitch [ Loop ] = 1 ;
      NewSocket = & UDPSocketArray [ Loop ] ;
      NewSocket -> NextSocketLink = NIL ;
      NewSocket -> UDPLayerInstanceLink = UDPLayerInstance ;
      NewSocket -> ListenService = NIL ;
      NewSocket -> ListenPortNo = AllocUDPPortN ( UDPLayerInstance ) ;
      return ( NewSocket ) ;
    } /* if */ ;
  } /* for */ ;
  return ( NIL ) ;
} /* AllocateUDPSocket */


void DeallocateUDPSocket ( PtrUDPSocketDescriptionRecT UDPSocket )
{
  PtrUDPLayerInstanceRecT UDPLayerInstance ;
  PtrUDPSocketDescriptionRecT TempSocket ;
  USHORT Loop ;
  /*-----------------------------------*/
  UDPLayerInstance = UDPSocket -> UDPLayerInstanceLink ;
  if ( UDPLayerInstance -> ListenSocketTopEntry )
  {
    TempSocket = UDPLayerInstance -> ListenSocketTopEntry ;
    if ( TempSocket == UDPSocket )
    {
      UDPLayerInstance -> ListenSocketTopEntry = UDPSocket -> NextSocketLink ;
    } /* if ... */
    else
    {
      while ( TempSocket )
      {
        if ( TempSocket -> NextSocketLink == UDPSocket )
        {
          TempSocket -> NextSocketLink = UDPSocket -> NextSocketLink ;
          break ;
        } /* if */ ;
        TempSocket = TempSocket -> NextSocketLink ;
      } /* while */ ;
    } /* if ... else */ ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < SocketArraySize ; Loop ++ )
  {
    TempSocket = & UDPSocketArray [ Loop ] ;
    if ( UDPSocket == TempSocket )
    {
      UDPSocketSwitch [ Loop ] = FALSE ;
      break ;
    } /* if */ ;
  } /* for */ ;
} /* DeallocateUDPSocket */


static PtrUDPSocketDescriptionRecT LocateUDPListen ( PtrUDPLayerInstanceRecT UDPLayerInstance ,
                                                     USHORT                  UDPPortNo )
{
  PtrUDPSocketDescriptionRecT TempSocket ;
  /*-----------------------------------*/
  TempSocket = UDPLayerInstance -> ListenSocketTopEntry ;
  while ( TempSocket )
  {
    if ( TempSocket -> ListenPortNo == UDPPortNo )
      break ;
    TempSocket = TempSocket -> NextSocketLink ;
  } /* if */ ;
  return ( TempSocket ) ;
} /* LocateUDPListen */


static void ProcessUDPPacked ( PtrIPLayerInstanceRecT IPLayerInstance ,
                               IPHeaderRecT *         IPHeader ,
                               PtrOSFrameRecT         Frame )
{
  USHORT SourcePortNo ;
  USHORT DestinationPortNo ;
  USHORT UDPPackedSize ;
  USHORT UDPCheckSum ;
  USHORT TestUDPCheckSum ;
  PtrUDPLayerInstanceRecT UDPLayerInstance ;
  PtrUDPSocketDescriptionRecT UDPSocket ;
  SocketUDPRecvEventServProcT UDPDataConsumer ;
  USHORT * CheckSumPointer ;
  void * UDPDataBuffer ;
  void * AllUDPDataBuffer ;
  /*-----------------------------------*/
  UDPLayerInstance = ( PtrUDPLayerInstanceRecT ) IPLayerInstance -> UDPInstanceEntry ;
  AllUDPDataBuffer = ( void * ) ( Frame -> Data + Frame -> Index ) ;
  TraceUDPHeader ( Frame ) ;
  SourcePortNo = ReadUSHORTSwapField ( Frame ) ;
  DestinationPortNo = ReadUSHORTSwapField ( Frame ) ;
  UDPPackedSize = ReadUSHORTSwapField ( Frame ) ;
  CheckSumPointer = ( USHORT * ) ( & Frame -> Data [ Frame -> Index ] ) ;
  UDPCheckSum = ReadUSHORTSwapField ( Frame ) ;
  if ( UDPPackedSize <= UDPHeaderSize )
    return ;
  UDPDataBuffer = ( void * ) ( Frame -> Data + Frame -> Index ) ;
  if ( UDPCheckSum )
  {
    * CheckSumPointer = 0 ;
    TestUDPCheckSum = CalculateUDPChecksum ( IPHeader -> SourceIPAddress , IPHeader -> DestinationIPAddress ,
                                             AllUDPDataBuffer , UDPPackedSize ) ;
    if ( ! TestUDPCheckSum )
      TestUDPCheckSum = 0xFFFF ;
    if ( UDPCheckSum != TestUDPCheckSum )
    {
      return ;
    } /* if */ ;
  } /* if */ ;
  UDPSocket = LocateUDPListen ( UDPLayerInstance , DestinationPortNo ) ;
  if ( UDPSocket )
  {
    UDPDataConsumer = UDPSocket -> ListenService ;
    if ( UDPDataConsumer )
      UDPDataConsumer ( UDPSocket , IPHeader -> SourceIPAddress , SourcePortNo ,
                        IPHeader -> DestinationIPAddress , DestinationPortNo ,
                        UDPDataBuffer , UDPPackedSize - UDPHeaderSize ) ;
  } /* if ... */
  else
  {
    if ( UDPLayerInstance -> DefaultListenService )
      UDPLayerInstance -> DefaultListenService ( IPHeader -> SourceIPAddress , SourcePortNo ,
                                                 IPHeader -> DestinationIPAddress , DestinationPortNo ,
                                                 UDPDataBuffer , UDPPackedSize - UDPHeaderSize ) ;
  } /* if ... else */ ;
} /* ProcessUDPPacked */


static void KillUDPData ( ULONG   SenderIPAddress ,
                          USHORT  SenderPortNo ,
                          ULONG   DestIPAddress ,
                          USHORT  DestPortNo ,
                          void *  UDPDataBuffer ,
                          USHORT  UDPDataSize )
{
} /* KillUDPData */


USHORT BindUDPSocket ( PtrUDPSocketDescriptionRecT UDPSocket ,
                       USHORT                      PortNo )
{
  if ( UDPSocket -> ListenPortNo == PortNo )
  {
    return ( UDP_BindPortOK ) ;
  } /* if */ ;
  if ( CheckUnique ( UDPSocket -> UDPLayerInstanceLink , PortNo ) )
  {
    UDPSocket -> ListenPortNo = PortNo ;
    return ( UDP_BindPortOK ) ;
  } /* if ... */
  else
  {
    return ( UDP_BindPortDupl ) ;
  } /* if ... else */ ;
} /* BindUDPSocket */


USHORT UDPSendData ( PtrUDPSocketDescriptionRecT Socket ,
                     ULONG                       DestIPAddress ,
                     USHORT                      DestPortNo ,
                     void *                      DataBufferPtr ,
                     USHORT                      DataSize )
{
  PtrUDPLayerInstanceRecT UDPLayerInstance ;
  PtrIPLayerInstanceRecT IPLayerInstance ;
  EMACInstanceRecT * InterfInstance ;
  void * AllUDPDataBuffer ;
  ULONG TargetAddress ;
  PtrOSFrameRecT OutFrame ;
  USHORT CheckSum ;
  USHORT BCMode ;
  USHORT Loop ;
  USHORT CheckSumOffset ;
  /*-----------------------------------*/
  if ( Socket )
  {
    if ( ! DataSize )
      return ( UDP_SendFailed ) ;
    UDPLayerInstance = Socket -> UDPLayerInstanceLink ;
    if ( DataSize > MaxUDPSegmentSize )
      return ( UDP_SendSegmentToBig ) ;
    IPLayerInstance = UDPLayerInstance -> IPLayerInstance ;
    InterfInstance = IPLayerInstance -> InterfaceInstance ;
    BCMode = FALSE ;
    if ( DestIPAddress == 0xFFFFFFFF )
    {
      BCMode = TRUE ;
      TargetAddress = DestIPAddress ;
    } /* if ... */
    else
    {
      if ( ( DestIPAddress & IPLayerInstance -> MyLocalNetMask ) == ( IPLayerInstance -> MyLocalIPAddress & IPLayerInstance -> MyLocalNetMask ) )
      {
        TargetAddress = DestIPAddress & ( ~ IPLayerInstance -> MyLocalNetMask ) ;
        if ( TargetAddress == ( ~ IPLayerInstance -> MyLocalNetMask ) )
          BCMode = TRUE ;
        TargetAddress = DestIPAddress ;
      } /* if ... */
      else
      {
        TargetAddress = IPLayerInstance-> MyDefaultGWIPAddress ;
        if ( ! TargetAddress )
        {
          return ( UDP_SendRouteFailed ) ;
        } /* if */ ;
      } /* if ... else */ ;
    } /* if ... else */ ;
    OutFrame = InterfInstance -> InterfaceAllocFrameMemory ( sizeof ( OSFrameRecT ) ) ;
    if ( OutFrame )
    {
      StartEthFrame ( OutFrame , NIL , InterfInstance -> InterfaceMACAddress , FRAME_IP ) ;
      CreateIPHeader ( OutFrame , Prot_UDP , IPLayerInstance -> MyLocalIPAddress , DestIPAddress ) ;
      PutUSHORTSwapField ( OutFrame , IPTotalLengthOffset , IPHeaderSize + UDPHeaderSize + DataSize ) ;
      CheckSum = CalculateChecksum ( & OutFrame -> Data [ IPVersOffset ] , IPHeaderSize ) ;
      PutUSHORTNoSwapField ( OutFrame , IPHeaderChecksumOffset , CheckSum ) ;
      AllUDPDataBuffer = ( void * ) ( OutFrame -> Data + OutFrame -> Length ) ;
      AddUSHORTSwapField ( OutFrame , Socket -> ListenPortNo ) ;
      AddUSHORTSwapField ( OutFrame , DestPortNo ) ;
      AddUSHORTSwapField ( OutFrame , DataSize + UDPHeaderSize ) ;
      CheckSumOffset = OutFrame -> Length ;
      AddUSHORTSwapField ( OutFrame , 0 ) ;   /* checksum */
      AddByteArrayField ( OutFrame , DataBufferPtr , DataSize ) ;
      CheckSum = CalculateUDPChecksum ( IPLayerInstance -> MyLocalIPAddress , DestIPAddress ,
                                        AllUDPDataBuffer , 
                                        OutFrame -> Length - EtherPackedHeaderRecTSize - IPHeaderSize ) ;
      if ( ! CheckSum )
        CheckSum = 0xFFFF ;
      PutUSHORTSwapField ( OutFrame , CheckSumOffset , CheckSum ) ;
      if ( BCMode )
      {
        for ( Loop = 0 ; Loop < EthMACAddressSize ; Loop ++ )
          OutFrame -> Data [ Loop ] = 0xFF ;
        InterfInstance -> InterfaceSendFrame ( OutFrame ) ;
      } /* if ... */
      else
      {
        IPLayerInstance -> SendIPPackedService ( IPLayerInstance , OutFrame , TargetAddress ) ;
      } /* if ... else */ ;
      InterfInstance -> InterfaceDeallocFrameMemory ( OutFrame ) ;
      return ( UDP_SendOK ) ;
    } /* if */ ;
  } /* if */ ;
  return ( UDP_SendFailed ) ;
} /* UDPSendData */


void SetDefaultUDPListen ( PtrUDPLayerInstanceRecT  UDPLayerInstance ,
                           AnyUDPRecvEventServProcT DefaultUDPDataConsumer )
{
  UDPLayerInstance -> DefaultListenService = DefaultUDPDataConsumer ;
} /* SetDefaultUDPListen */


PtrUDPSocketDescriptionRecT CreateUDPSocket ( PtrUDPLayerInstanceRecT UDPLayerInstance )
{
  PtrUDPSocketDescriptionRecT NewSocket ;
  /*-----------------------------------*/
  NewSocket = AllocateUDPSocket ( UDPLayerInstance ) ;
  return ( NewSocket ) ;
} /* CreateUDPSocket */


void SetUDPListen ( PtrUDPSocketDescriptionRecT UDPSocket ,
                    SocketUDPRecvEventServProcT UDPDataConsumer )
{
  PtrUDPLayerInstanceRecT UDPLayerInstance ;
  /*-----------------------------------*/
  UDPLayerInstance = UDPSocket -> UDPLayerInstanceLink ;
  UDPSocket -> ListenService = UDPDataConsumer ;
  UDPSocket -> NextSocketLink = UDPLayerInstance -> ListenSocketTopEntry ;
  UDPLayerInstance -> ListenSocketTopEntry = UDPSocket ;
} /* SetUDPListen */


void UDPInitInstance ( PtrUDPLayerInstanceRecT UDPLayerInstance ,
                       PtrIPLayerInstanceRecT  IPLayerInstance )
{
  USHORT Loop ;
  /*-----------------------------------*/
  UDPLayerInstance -> IPLayerInstance = IPLayerInstance ;
  UDPLayerInstance -> DefaultListenService = KillUDPData ;
  UDPLayerInstance -> ListenSocketTopEntry = NIL ;
  UDPLayerInstance -> LastLocalPort = AutoIncrementUDPPortNoLoBound ;
  IPLayerInstance -> UDPProcessService = ProcessUDPPacked ;
  IPLayerInstance -> UDPInstanceEntry = UDPLayerInstance ;
  for ( Loop = 0 ; Loop < SocketArraySize ; Loop ++ )
    UDPSocketSwitch [ Loop ] = 0 ;
} /* UDPInitInstance */


PtrUDPLayerInstanceRecT GetUDPInstance ( PtrUDPSocketDescriptionRecT UDPSocket )
{
  return ( UDPSocket -> UDPLayerInstanceLink ) ;
} /* GetUDPInstance */


PtrIPLayerInstanceRecT GetIPInstanceFromUDPInstance ( PtrUDPLayerInstanceRecT UDPLayerInstance )
{
  return ( UDPLayerInstance -> IPLayerInstance ) ;
} /* GetIPInstanceFromUDPInstance */


static UDPLayerInstanceRecT LocalUDPLayerInstance ;
PtrUDPLayerInstanceRecT StandardAddUDPInstance ( PtrIPLayerInstanceRecT IPLayerInstance )
{
  PtrUDPLayerInstanceRecT UDPLayerInstance ;
  /*-----------------------------------*/
  UDPLayerInstance = & LocalUDPLayerInstance ;
  UDPInitInstance ( UDPLayerInstance , IPLayerInstance ) ;
  return ( UDPLayerInstance ) ;
} /* StandardAddUDPInstance */

